#include "Comm.h"

// Initialize communication pins and Serial1
void initComm() {
    // Set up control pins
    pinMode(DTR_PIN, OUTPUT);
    pinMode(RTS_PIN, OUTPUT);
    pinMode(CTS_PIN, INPUT);
    
    // Initial pin states
    digitalWrite(DTR_PIN, HIGH);  // DTR inactive (HIGH = inactive for RS232)
    digitalWrite(RTS_PIN, HIGH);  // RTS inactive
    
    // Configure serial port based on settings
    configureSerial();
}

// Configure Serial1 based on current settings
void configureSerial() {
    long baudRate = BAUD_RATES[serialSettings.baudRateIndex];
    int dataBits = DATA_BITS_OPTIONS[serialSettings.dataBitsIndex];
    
    // Determine Arduino serial config based on settings
    uint8_t config = 0;
    
    // Set appropriate serial configuration
    if (dataBits == 8) {
        // 8 data bits configurations
        switch (serialSettings.parityIndex) {
            case PARITY_NONE:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_8N1 : SERIAL_8N2;
                break;
            case PARITY_ODD:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_8O1 : SERIAL_8O2;
                break;
            case PARITY_EVEN:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_8E1 : SERIAL_8E2;
                break;
        }
    } else { // 7 data bits
        // 7 data bits configurations
        switch (serialSettings.parityIndex) {
            case PARITY_NONE:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_7N1 : SERIAL_7N2;
                break;
            case PARITY_ODD:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_7O1 : SERIAL_7O2;
                break;
            case PARITY_EVEN:
                config = serialSettings.stopBitsIndex == 0 ? SERIAL_7E1 : SERIAL_7E2;
                break;
        }
    }
    
    // Begin Serial1 with the determined settings
    Serial1.begin(baudRate, config);
    
    // Flush any existing data
    Serial1.flush();
}

// Set DTR pin state
void setDTR(bool state) {
    // RS232 signals are inverted (LOW = active, HIGH = inactive)
    digitalWrite(DTR_PIN, state ? LOW : HIGH);
}

// Set RTS pin state
void setRTS(bool state) {
    // RS232 signals are inverted (LOW = active, HIGH = inactive)
    digitalWrite(RTS_PIN, state ? LOW : HIGH);
}

// Get CTS pin state
bool getCTS() {
    // RS232 signals are inverted (LOW = active, HIGH = inactive)
    return digitalRead(CTS_PIN) == LOW;
}

// Wait for CTS to become active with timeout
bool waitForCTS(unsigned long timeout) {
    if (!serialSettings.flowControlEnabled) {
        return true;  // If flow control is disabled, always return true
    }
    
    unsigned long startTime = millis();
    while (!getCTS()) {
        // Check for timeout
        if (millis() - startTime > timeout) {
            return false;  // Timeout waiting for CTS
        }
        
        // Check for user cancellation
        if (getKey() == KEY_ESCAPE) {
            return false;
        }
        
        delay(1);
    }
    return true;  // CTS became active
}

// Send a file over Serial1
TransferStatus sendFile(const char* filename) {
    // Calculate progress update interval
    unsigned long progressInterval = 250; // Update progress every 250 milliseconds
    unsigned long lastProgressUpdate = 0;
    
    // Open the file
    clearLCD();
    printLCD(0, 0, "Send CNC Program");
    printLCD(0, 1, filename);
    
    // Raise DTR to indicate active connection
    setDTR(true);
    
    // Wait for key press to start sending
    printLCD(0, 2, "Press any key");
    while (getKey() == NO_KEY) {
        delay(50);
    }
    
    // Open the file after user confirmation
    File dataFile = SD.open(filename, FILE_READ);
    if (!dataFile) {
        setDTR(false);
        printLCD(0, 2, "File not found!");
        delay(2000);
        return TRANSFER_FILE_ERROR;
    }
    
    // Initialize counters
    unsigned long bytesSent = 0;
    unsigned long fileSize = dataFile.size();
    unsigned long sendStartTime = millis();
    unsigned long flowCheckCounter = 0;
    
    // Clear any existing serial data
    while (Serial1.available()) {
        Serial1.read();
    }
    
    // Set RTS if flow control is enabled
    if (serialSettings.flowControlEnabled) {
        setRTS(true);
    }
    
    // Display initial progress
    printLCD(0, 2, "Sending: 0/" + String(fileSize));
    printLCD(0, 3, "Press E to cancel");
    
    // Read and send the file
    uint8_t buffer[COMM_BUFFER_SIZE];
    int bytesRead;
    bool cancelled = false;
    
    while ((bytesRead = dataFile.read(buffer, sizeof(buffer))) > 0 && !cancelled) {
        // Process buffer in smaller chunks for better flow control
        for (int i = 0; i < bytesRead; i++) {
            // Check for cancellation periodically
            if (i % 10 == 0 && getKey() == KEY_ESCAPE) {
                cancelled = true;
                break;
            }
            
            // Check flow control every 10 bytes if enabled
            flowCheckCounter++;
            if (serialSettings.flowControlEnabled && (flowCheckCounter % 10 == 0)) {
                if (!waitForCTS(5000)) {  // 5 second timeout for CTS
                    dataFile.close();
                    setDTR(false);
                    setRTS(false);
                    printLCD(0, 2, "Flow control timeout!");
                    delay(2000);
                    return TRANSFER_TIMEOUT;
                }
            }
            
            // Send the byte
            Serial1.write(buffer[i]);
            bytesSent++;
            
            // Update progress display periodically
            unsigned long currentTime = millis();
            if (currentTime - lastProgressUpdate > progressInterval) {
                // Calculate percentage
                int progress = (bytesSent * 100) / fileSize;
                
                // Update LCD with current progress
                printLCD(0, 2, "Sent: " + String(bytesSent) + "/" + String(fileSize) + "   ");
                
                // Display progress bar
                String progressBar = "[";
                for (int p = 0; p < 18; p++) {
                    progressBar += (p < progress * 18 / 100) ? "#" : " ";
                }
                progressBar += "]";
                printLCD(0, 3, progressBar);
                
                lastProgressUpdate = currentTime;
            }
        }
    }
    
    // Handle cancellation
    if (cancelled) {
        dataFile.close();
        setDTR(false);
        if (serialSettings.flowControlEnabled) {
            setRTS(false);
        }
        printLCD(0, 2, "Send cancelled!");
        delay(2000);
        return TRANSFER_CANCELLED;
    }
    
    // Ensure all data is sent
    Serial1.flush();
    
    // Close the file
    dataFile.close();
    
    // Lower DTR and RTS when done
    setDTR(false);
    if (serialSettings.flowControlEnabled) {
        setRTS(false);
    }
    
    // Calculate transfer time and rate
    unsigned long transferTime = millis() - sendStartTime;
    float transferRate = (bytesSent * 1000.0) / transferTime;  // bytes per second
    
    // Display completion information
    clearLCD();
    printLCD(0, 0, "Send Complete");
    printLCD(0, 1, "Bytes: " + String(bytesSent));
    printLCD(0, 2, "Time: " + String(transferTime / 1000.0, 1) + "s");
    printLCD(0, 3, "Rate: " + String(transferRate, 0) + " B/s");
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(50);
    }
    
    return TRANSFER_SUCCESS;
}

// Receive a file over Serial1 with reduced memory impact
TransferStatus receiveFile(const char* filename) {
    // Pre-allocate progress strings to avoid memory fragmentation
    char progressStr[21]; // For LCD display
    
    clearLCD();
    printLCD(0, 0, "Receive Program");
    printLCD(0, 1, filename);
    
    // Ensure serial is properly configured with correct settings
    configureSerial();
    
    // Raise DTR to indicate active connection
    setDTR(true);
    
    // Wait for key press to start receiving
    printLCD(0, 2, "Press any key to start");
    printLCD(0, 3, "receiving");
    while (getKey() == NO_KEY) {
        delay(50);
    }
    
    // Delete existing file to ensure clean write
    if (SD.exists(filename)) {
        SD.remove(filename);
    }
    
    // Open file in binary write mode
    File dataFile = SD.open(filename, FILE_WRITE);
    if (!dataFile) {
        setDTR(false);
        printLCD(0, 2, "Cannot create file!");
        delay(2000);
        return TRANSFER_FILE_ERROR;
    }
    
    // Clear any existing data in the serial buffer
    while (Serial1.available()) {
        Serial1.read();
    }
    
    // Initialize counters and timers
    uint32_t bytesReceived = 0;
    uint32_t receiveStartTime = millis();
    uint32_t lastActivity = receiveStartTime;
    uint32_t lastProgressUpdate = 0;
    uint32_t receiveTimeout = 5000; // 5 seconds without data is timeout
    
    // Set RTS if flow control is enabled, indicating we're ready to receive
    if (serialSettings.flowControlEnabled) {
        setRTS(true);
    }
    
    // Display initial progress - using snprintf to prevent buffer issues
    snprintf(progressStr, sizeof(progressStr), "Receiving: 0 bytes");
    printLCD(0, 2, progressStr);
    printLCD(0, 3, "Press E to stop");
    
    // Receive loop
    bool receiving = true;
    bool endDetected = false;
    uint8_t recentByte = 0;
    
    while (receiving && !endDetected) {
        // Check for user cancellation (only occasionally to avoid slowing down)
        if (bytesReceived % 128 == 0) {
            char key = getKey();
            if (key == KEY_ESCAPE) {
                // Ensure file is properly closed
                dataFile.flush();
                dataFile.close();
                
                setDTR(false);
                if (serialSettings.flowControlEnabled) {
                    setRTS(false);
                }
                
                printLCD(0, 2, "Receive stopped");
                delay(2000);
                return TRANSFER_CANCELLED;
            }
        }
        
        // Check for data reception timeout
        if (millis() - lastActivity > receiveTimeout && bytesReceived > 0) {
            // End detection based on timeout with data
            endDetected = true;
        }
        
        // Process available data
        while (Serial1.available() > 0) {
            // Reset activity timer
            lastActivity = millis();
            
            // Get the byte
            uint8_t inByte = Serial1.read();
            recentByte = inByte;  // Save for debugging
            
            // Write the byte to file (no delay between writes)
            if (!dataFile.write(inByte)) {
                // If write fails, close file and return error
                dataFile.close();
                setDTR(false);
                if (serialSettings.flowControlEnabled) {
                    setRTS(false);
                }
                printLCD(0, 2, "Write error!");
                delay(2000);
                return TRANSFER_ERROR;
            }
            
            bytesReceived++;
            
            // Toggle RTS for flow control only if needed (and not too frequently)
            if (serialSettings.flowControlEnabled && bytesReceived % 64 == 0) {
                setRTS(false);
                setRTS(true);
            }
            
            // Flush every 512 bytes to match SD card sector size
            if (bytesReceived % 512 == 0) {
                dataFile.flush();
            }
            
            // Update progress display periodically (without using String class)
            uint32_t currentTime = millis();
            if (currentTime - lastProgressUpdate > 250) {
                // Use snprintf instead of String to avoid memory fragmentation
                snprintf(progressStr, sizeof(progressStr), "Rcvd: %lu bytes", bytesReceived);
                printLCD(0, 2, progressStr);
                
                // Show last byte as hex (for debugging)
                snprintf(progressStr, sizeof(progressStr), "Last: %02X", recentByte);
                printLCD(0, 3, progressStr);
                
                lastProgressUpdate = currentTime;
            }
        }
        
        // Minimal delay to prevent CPU hogging
        if (Serial1.available() == 0) {
            // Only delay if no data is available
            delay(1);
        }
    }
    
    // Ensure the file is properly closed
    dataFile.flush();
    dataFile.close();
    
    // Lower DTR and RTS when done
    setDTR(false);
    if (serialSettings.flowControlEnabled) {
        setRTS(false);
    }
    
    // Calculate transfer time and rate
    uint32_t transferTime = millis() - receiveStartTime;
    float transferRate = (bytesReceived * 1000.0f) / max(transferTime, 1UL);  // bytes per second
    
    // Display completion information
    // Display completion information using dtostrf for floating point values
    clearLCD();
    printLCD(0, 0, "Receive Complete");
    
    // Bytes received (integer value)
    snprintf(progressStr, sizeof(progressStr), "Bytes: %lu", bytesReceived);
    printLCD(0, 1, progressStr);
    
    // Time in seconds - use dtostrf for float conversion
    char floatStr[10];
    float timeSeconds = transferTime / 1000.0f;
    dtostrf(timeSeconds, 4, 1, floatStr);  // Format: width 4, 1 decimal place
    snprintf(progressStr, sizeof(progressStr), "Time: %ss", floatStr);
    printLCD(0, 2, progressStr);
    
    // Transfer rate - use dtostrf again
    dtostrf(transferRate, 4, 0, floatStr);  // Format: width 4, 0 decimal places
    snprintf(progressStr, sizeof(progressStr), "Rate: %s B/s", floatStr);
    printLCD(0, 3, progressStr);
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(50);
    }
    
    return TRANSFER_SUCCESS;
}