#include "common.h"

// Baud rate options
const long BAUD_RATES[BAUD_COUNT] = {110, 300, 1200, 2400, 4800, 9600, 19200, 28800, 38400, 57600, 115200};
const char* BAUD_NAMES[BAUD_COUNT] = {"110", "300", "1200", "2400", "4800", "9600", "19200", "28800", "38400", "57600", "115200"};

// Data bits options
const int DATA_BITS_OPTIONS[DATA_BITS_COUNT] = {7, 8};

// Parity names
const char* PARITY_NAMES[PARITY_COUNT] = {"None", "Odd", "Even"};

// Stop bits options
const float STOP_BITS_OPTIONS[STOP_BITS_COUNT] = {1.0, 2.0};
const char* STOP_BITS_NAMES[STOP_BITS_COUNT] = {"1", "2"};

// Global settings variable
SerialSettings serialSettings;

// Load settings from EEPROM
void loadSettingsFromEEPROM() {
    // Read settings structure from EEPROM
    EEPROM.get(EEPROM_SETTINGS_ADDR, serialSettings);
    
    // Check if the settings are valid (using a flag)
    if (serialSettings.validFlag != EEPROM_VALID_FLAG) {
        // If not valid, set defaults
        setDefaultSettings();
        saveSettingsToEEPROM();
    }
}

// Save settings to EEPROM
void saveSettingsToEEPROM() {
    // Set the valid flag
    serialSettings.validFlag = EEPROM_VALID_FLAG;
    
    // Write the entire structure to EEPROM
    EEPROM.put(EEPROM_SETTINGS_ADDR, serialSettings);
}

// Set default settings
void setDefaultSettings() {
    serialSettings.baudRateIndex = 5;        // 9600 baud
    serialSettings.dataBitsIndex = 1;        // 8 data bits
    serialSettings.parityIndex = PARITY_NONE; // No parity
    serialSettings.stopBitsIndex = 0;        // 1 stop bit
    serialSettings.flowControlEnabled = false; // No flow control
    strcpy(serialSettings.currentFileName, ""); // No file
}

// Initialize SD card
bool initSDCard() {
    clearLCD();
    printLCD(0, 0, "SD Card");
    printLCD(0, 1, "Initializing...");
    
    // Initialize SD card
    if (!SD.begin(SD_CS_PIN)) {
        printLCD(0, 2, "SD init failed!");
        delay(2000);
        return false;
    }
    
    printLCD(0, 2, "SD Card ready");
    delay(1000);
    return true;
}

// Select a file from the SD card
// Select a file from the SD card
String selectFileFromSD() {
    // Initialize SD card if needed
    if (!SD.begin(SD_CS_PIN)) {
        clearLCD();
        printLCD(0, 0, "SD Card Error");
        printLCD(0, 1, "Cannot initialize");
        printLCD(0, 3, "Any key to return");
        
        while (getKey() == NO_KEY) {
            delay(100);
        }
        return "";
    }
    
    // Open root directory
    File root = SD.open("/");
    if (!root) {
        clearLCD();
        printLCD(0, 0, "SD Card Error");
        printLCD(0, 1, "Cannot open root");
        printLCD(0, 3, "Any key to return");
        
        while (getKey() == NO_KEY) {
            delay(100);
        }
        return "";
    }
    
    clearLCD();
    printLCD(0, 0, "Reading files...");
    
    // Array to store filenames
    #define MAX_FILES 50
    String fileNames[MAX_FILES];
    int fileCount = 0;
    
    // Read all filenames in root directory
    while (fileCount < MAX_FILES) {
        File entry = root.openNextFile();
        if (!entry) {
            // No more files
            break;
        }
        
        // Skip directories
        if (!entry.isDirectory()) {
            fileNames[fileCount] = String(entry.name());
            fileCount++;
        }
        
        entry.close();
    }
    
    root.close();
    
    // No files found
    if (fileCount == 0) {
        clearLCD();
        printLCD(0, 0, "SD Card");
        printLCD(0, 1, "No files found");
        printLCD(0, 3, "Any key to return");
        
        while (getKey() == NO_KEY) {
            delay(100);
        }
        return "";
    }
    
    // File browser interface
    int selectedIndex = 0;
    int displayOffset = 0;
    char key;
    bool needsRefresh = true;  // Flag to control screen refresh
    
    while (true) {
        // Only refresh the display when needed
        if (needsRefresh) {
            clearLCD();
            printLCD(0, 0, "SD Files (" + String(fileCount) + ")");
            
            // Display up to 3 files
            for (int i = 0; i < 3; i++) {
                int fileIndex = displayOffset + i;
                if (fileIndex < fileCount) {
                    String prefix = (fileIndex == selectedIndex) ? "> " : "  ";
                    String displayName = fileNames[fileIndex];
                    if (displayName.length() > 18) {
                        displayName = displayName.substring(0, 15) + "...";
                    }
                    printLCD(0, i + 1, prefix + displayName);
                } else {
                    printLCD(0, i + 1, "                    "); // Clear line
                }
            }
            
            needsRefresh = false;  // Reset the flag after refresh
        }
        
        // Wait for key press
        key = getKey();
        
        if (key == KEY_UP) {
            if (selectedIndex > 0) {
                selectedIndex--;
                if (selectedIndex < displayOffset) {
                    displayOffset = selectedIndex;
                }
                needsRefresh = true;  // Need to refresh after navigation
            }
        } else if (key == KEY_DOWN) {
            if (selectedIndex < fileCount - 1) {
                selectedIndex++;
                if (selectedIndex >= displayOffset + 3) {
                    displayOffset = selectedIndex - 2;
                }
                needsRefresh = true;  // Need to refresh after navigation
            }
        } else if (key == KEY_ENTER) {
            // File selected, return the filename
            String selectedFile = fileNames[selectedIndex];
            saveCurrentFileName(selectedFile);
            return selectedFile;
        } else if (key == KEY_ESCAPE) {
            // Cancel selection
            return "";
        }
        
        delay(100);
    }
}

// Map for translating keypad numbers to characters
const char* keypadChars[] = {
    " 0",          // 0
    ".1",          // 1
    "ABC2",        // 2
    "DEF3",        // 3
    "GHI4",        // 4
    "JKL5",        // 5
    "MNO6",        // 6
    "PQRS7",       // 7
    "TUV8",        // 8
    "WXYZ9"        // 9
};

// Input a filename using the keypad (8.3 format)
String inputFileName() {
    clearLCD();
    printLCD(0, 0, "Enter Filename");
    printLCD(0, 3, "O:Save B:Back E:Exit");
    
    String filename = "";
    bool insertPeriod = false;
    int charPosition = 0;
    int lastKey = -1;
    int keyPressCount = 0;
    unsigned long lastKeyPressTime = 0;
    bool namePart = true; // true = name part, false = extension part
    
    while (true) {
        // Display current filename
        printLCD(0, 1, filename + "_       ");
        
        // Character selection hint
        if (lastKey >= 0 && lastKey <= 9) {
            int charIndex = keyPressCount % strlen(keypadChars[lastKey]);
            String availChars = String(keypadChars[lastKey]);
            printLCD(0, 2, "Select: " + String(availChars[charIndex]) + "   ");
        } else {
            printLCD(0, 2, "                    ");
        }
        
        char key = getKey();
        
        // Process the key
        if (key != NO_KEY) {
            unsigned long currentTime = millis();
            
            // Reset character selection if different key pressed
            if (key >= '0' && key <= '9') {
                int numericKey = key - '0';
                
                // Check if it's the same key as last time
                if (numericKey == lastKey && (currentTime - lastKeyPressTime) < 1000) {
                    // Same key, increment character position
                    keyPressCount++;
                } else {
                    // Different key, reset character position
                    lastKey = numericKey;
                    keyPressCount = 0;
                }
                
                lastKeyPressTime = currentTime;
            } else if (key == KEY_RIGHT) {
                // Confirm character selection and add to filename
                if (lastKey >= 0 && lastKey <= 9 && strlen(keypadChars[lastKey]) > 0) {
                    int charIndex = keyPressCount % strlen(keypadChars[lastKey]);
                    char selectedChar = keypadChars[lastKey][charIndex];
                    
                    // Check filename length constraints
                    if (namePart && charPosition < 8) {
                        filename += selectedChar;
                        charPosition++;
                    } else if (!namePart && charPosition < 12) { // 8+1+3
                        filename += selectedChar;
                        charPosition++;
                    }
                    
                    lastKey = -1;
                    keyPressCount = 0;
                }
            } else if (key == KEY_LEFT || key == KEY_BACKSPACE) {
                // Backspace
                if (filename.length() > 0) {
                    // If deleting period, switch back to name part
                    if (filename.charAt(filename.length() - 1) == '.') {
                        namePart = true;
                    }
                    
                    filename = filename.substring(0, filename.length() - 1);
                    charPosition--;
                    lastKey = -1;
                    keyPressCount = 0;
                }
            } else if (key == '#' && namePart && filename.length() > 0 && filename.indexOf('.') == -1) {
                // Add period (only in name part, and only if there isn't one already)
                filename += '.';
                charPosition++;
                namePart = false;
                lastKey = -1;
                keyPressCount = 0;
            } else if (key == KEY_ENTER) {
                // Save filename
                if (filename.length() > 0) {
                    // Ensure there's an extension if there's a name
                    if (filename.indexOf('.') == -1) {
                        filename += ".NC"; // Default extension
                    }
                    saveCurrentFileName(filename);
                    return filename;
                }
            } else if (key == KEY_ESCAPE) {
                // Cancel
                return "";
            }
        }
        
        delay(100);
    }
}

// Save current filename to settings
void saveCurrentFileName(const String &fileName) {
    if (fileName.length() < sizeof(serialSettings.currentFileName)) {
        strcpy(serialSettings.currentFileName, fileName.c_str());
        saveSettingsToEEPROM();
    }
}