#include "Menus.h"
#include "Common.h"
#include "Comm.h"

#define MAX_MENUS 3

// Global menu variables
Menu menus[MAX_MENUS];
uint8_t currentMenuId = 0;
uint8_t selectedItem = 0;
uint8_t menuDisplayOffset = 0;

// Initialize the menu system
void initMenuSystem() {
    // Register main menu
    registerMenu(0, "NCComm Main Menu", 0);
    addMenuItem(0, "Send CNC Program", sendCNCProgram, 0);
    addMenuItem(0, "Receive Program", receiveProgram, 0);
    addMenuItem(0, "Set File", browseSDFiles, 0);
    addMenuItem(0, "RS232 Settings", NULL, 1);
    addMenuItem(0, "About...", showAboutInfo, 0);
    
    // Register RS232 settings submenu
    registerMenu(1, "RS232 Settings", 0);
    addMenuItem(1, "Baud Rate", setBaudRate, 0);
    addMenuItem(1, "Data Bits", setDataBits, 0);
    addMenuItem(1, "Parity", setParity, 0);
    addMenuItem(1, "Stop Bits", setStopBits, 0);
    addMenuItem(1, "Flow Control", setFlowControl, 0);  // Added RTS/CTS option
    addMenuItem(1, "Save Settings", saveSerialSettings, 0);
    
    // Start with main menu
    navigateToMenu(0);
}

// Register a menu
void registerMenu(uint8_t menuId, const String &title, uint8_t parentMenuId) {
    if (menuId < MAX_MENUS) {
        // Create a fresh String object from a const char*
        menus[menuId].title = String(title.c_str());
        menus[menuId].itemCount = 0;
        menus[menuId].parentMenuId = parentMenuId;
    }
}

// Add a menu item to a menu
void addMenuItem(uint8_t menuId, const String &title, void (*action)(), uint8_t subMenuId) {
    if (menuId < MAX_MENUS && menus[menuId].itemCount < MAX_MENU_ITEMS) {
        MenuItem item;
        // Create a fresh String object from a const char*
        item.title = String(title.c_str());
        item.action = action;
        item.subMenuId = subMenuId;
        menus[menuId].items[menus[menuId].itemCount++] = item;
    }
}

// Display the current menu
void displayMenu() {
    clearLCD();
    
    // Display menu title
    printLCD(0, 0, menus[currentMenuId].title);
    
    // Display visible menu items (up to 3 items for a 4-line display)
    uint8_t displayCount = min(3, menus[currentMenuId].itemCount);
    for (uint8_t i = 0; i < displayCount; i++) {
        uint8_t itemIndex = menuDisplayOffset + i;
        if (itemIndex < menus[currentMenuId].itemCount) {
            String prefix = (itemIndex == selectedItem) ? "> " : "  ";
            printLCD(0, i + 1, prefix + menus[currentMenuId].items[itemIndex].title);
        }
    }
    
    // Display navigation hints if needed
    if (menus[currentMenuId].itemCount > 3) {
        if (menuDisplayOffset > 0) {
            printLCD(19, 1, "^");
        }
        if (menuDisplayOffset + 3 < menus[currentMenuId].itemCount) {
            printLCD(19, 3, "v");
        }
    }
}

// Handle menu navigation based on keypad input
void handleMenuNavigation() {
    char key = getKey();
    
    if (key != NO_KEY) {
        switch (key) {
            case KEY_UP:
                if (selectedItem > 0) {
                    selectedItem--;
                    if (selectedItem < menuDisplayOffset) {
                        menuDisplayOffset = selectedItem;
                    }
                    displayMenu();
                }
                break;
                
            case KEY_DOWN:
                if (selectedItem < menus[currentMenuId].itemCount - 1) {
                    selectedItem++;
                    if (selectedItem >= menuDisplayOffset + 3) {
                        menuDisplayOffset = selectedItem - 2;
                    }
                    displayMenu();
                }
                break;
                
            case KEY_ENTER:
            case KEY_RIGHT:
                executeMenuAction();
                break;
                
            case KEY_ESCAPE:
            case KEY_LEFT:
                if (currentMenuId > 0) {
                    navigateToMenu(menus[currentMenuId].parentMenuId);
                }
                break;
                
            case KEY_F1:  // F1 key for help
                showHelp();
                break;
                
            case KEY_F2:  // F2 key for status
                showQuickActions();
                break;
        }
    }
}

// Show help information based on current menu
void showHelp() {
    clearLCD();
    printLCD(0, 0, "Help: " + menus[currentMenuId].title);
    printLCD(0, 1, "U/D: Navigate");
    printLCD(0, 2, "O/R: Select");
    printLCD(0, 3, "E/L: Back");
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(100);
    }
    
    displayMenu();
}

// Show quick actions/status
void showQuickActions() {
    clearLCD();
    printLCD(0, 0, "NCComm Status");
    printLCD(0, 1, "File: No file set");
    printLCD(0, 2, "RS232: 9600,8,N,1");
    printLCD(0, 3, "Any key to return");
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(100);
    }
    
    displayMenu();
}

// Execute the selected menu item's action
void executeMenuAction() {
    MenuItem item = menus[currentMenuId].items[selectedItem];
    
    // Navigate to submenu if one is specified
    if (item.subMenuId > 0) {
        navigateToMenu(item.subMenuId);
        return;
    }
    
    // Execute action if one is specified
    if (item.action != NULL) {
        item.action();
    }
}

// Navigate to a specific menu
void navigateToMenu(uint8_t menuId) {
    if (menuId < MAX_MENUS) {
        currentMenuId = menuId;
        selectedItem = 0;
        menuDisplayOffset = 0;
        displayMenu();
    }
}

// CNC Operations Functions (empty implementations)

void sendCNCProgram() {
    clearLCD();
    printLCD(0, 0, "Send CNC Program");
    
    // Check if a file is already selected
    if (strlen(serialSettings.currentFileName) == 0) {
        printLCD(0, 1, "No file selected");
        printLCD(0, 2, "Choose a file first");
        delay(1500);
        
        // Let user browse and select a file
        String selectedFile = selectFileFromSD();
        
        // If user cancelled file selection, abort operation
        if (selectedFile.length() == 0) {
            clearLCD();
            printLCD(0, 0, "Send CNC Program");
            printLCD(0, 1, "Operation cancelled");
            printLCD(0, 2, "No file selected");
            delay(2000);
            displayMenu();
            return;
        }
    }
    
    // Initialize communication if not already done
    initComm();
    
    // Send the file and get the status
    TransferStatus status = sendFile(serialSettings.currentFileName);
    
    // Handle the status if needed (most messaging is handled in sendFile)
    switch (status) {
        case TRANSFER_SUCCESS:
            // Success already displayed in sendFile
            break;
            
        case TRANSFER_CANCELLED:
        case TRANSFER_FILE_ERROR:
        case TRANSFER_TIMEOUT:
        case TRANSFER_ERROR:
            // Error messages already displayed in sendFile
            break;
    }
    
    // Return to menu
    displayMenu();
}

void receiveProgram() {
    clearLCD();
    printLCD(0, 0, "Receive Program");
    
    // Check if a file is already selected
    if (strlen(serialSettings.currentFileName) == 0) {
        // No file selected, ask user to create a new file or browse existing
        printLCD(0, 1, "No file selected");
        printLCD(0, 2, "1:New file 2:Browse");
        
        // Wait for user choice
        char key = NO_KEY;
        while (key != '1' && key != '2' && key != KEY_ESCAPE) {
            key = getKey();
            if (key != NO_KEY) {
                if (key == KEY_ESCAPE) {
                    clearLCD();
                    printLCD(0, 0, "Receive Program");
                    printLCD(0, 1, "Operation cancelled");
                    delay(2000);
                    displayMenu();
                    return;
                }
            }
            delay(100);
        }
        
        String selectedFile = "";
        if (key == '1') {
            // Create new file
            selectedFile = inputFileName();
        } else if (key == '2') {
            // Browse existing files
            selectedFile = selectFileFromSD();
        }
        
        // If user cancelled file selection, abort operation
        if (selectedFile.length() == 0) {
            clearLCD();
            printLCD(0, 0, "Receive Program");
            printLCD(0, 1, "Operation cancelled");
            printLCD(0, 2, "No file selected");
            delay(2000);
            displayMenu();
            return;
        }
    }
    
    // Warn user if file exists - it will be overwritten
    if (SD.exists(serialSettings.currentFileName)) {
        clearLCD();
        printLCD(0, 0, "Warning!");
        printLCD(0, 1, "File exists:");
        printLCD(0, 2, serialSettings.currentFileName);
        printLCD(0, 3, "O:Overwrite E:Cancel");
        
        // Wait for confirmation
        char key = NO_KEY;
        while (key != KEY_ENTER && key != KEY_ESCAPE) {
            key = getKey();
            delay(100);
        }
        
        if (key == KEY_ESCAPE) {
            clearLCD();
            printLCD(0, 0, "Receive Program");
            printLCD(0, 1, "Operation cancelled");
            delay(2000);
            displayMenu();
            return;
        }
    }
    
    // Initialize communication if not already done
    initComm();
    
    // Receive to the file and get the status
    TransferStatus status = receiveFile(serialSettings.currentFileName);
    
    // Handle the status if needed (most messaging is handled in receiveFile)
    switch (status) {
        case TRANSFER_SUCCESS:
            // Success already displayed in receiveFile
            break;
            
        case TRANSFER_CANCELLED:
        case TRANSFER_FILE_ERROR:
        case TRANSFER_TIMEOUT:
        case TRANSFER_ERROR:
            // Error messages already displayed in receiveFile
            break;
    }
    
    // Return to menu
    displayMenu();
}

void browseSDFiles() {
    clearLCD();
    printLCD(0, 0, "File Operations");
    printLCD(0, 1, "1: Browse SD files");
    printLCD(0, 2, "2: Create new file");
    printLCD(0, 3, "E: Cancel");
    
    char key;
    String selectedFile = "";
    
    // Wait for option selection
    while (true) {
        key = getKey();
        
        if (key == '1') {
            // Browse existing files
            selectedFile = selectFileFromSD();
            break;
        } else if (key == '2') {
            // Create new filename
            selectedFile = inputFileName();
            break;
        } else if (key == KEY_ESCAPE) {
            // Cancel operation
            break;
        }
        
        delay(100);
    }
    
    // Display the result if a file was selected
    if (selectedFile.length() > 0) {
        clearLCD();
        printLCD(0, 0, "File Selected:");
        
        // If filename is too long, scroll it
        if (selectedFile.length() > 20) {
            printLCD(0, 1, selectedFile.substring(0, 20));
            printLCD(0, 2, selectedFile.substring(20));
        } else {
            printLCD(0, 1, selectedFile);
        }
        
        printLCD(0, 3, "Any key to continue");
        
        // Wait for key press
        while (getKey() == NO_KEY) {
            delay(100);
        }
    }
    
    displayMenu();
}

void showAboutInfo() {
    clearLCD();
    printLCD(0, 0, "NCComm System");
    printLCD(0, 1, "(C) 2025");
    printLCD(0, 2, "By: Steven Mackaay");
    printLCD(0, 3, "Any key to return");
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(100);
    }
    
    displayMenu();
}

// RS232 Settings Functions (empty implementations)

// RS232 Settings Functions

void setBaudRate() {
    clearLCD();
    printLCD(0, 0, "Set Baud Rate");
    printLCD(0, 1, "Use U/D to select");
    
    char key;
    uint8_t currentIndex = serialSettings.baudRateIndex;
    
    while (true) {
        // Display current selection
        printLCD(0, 2, "Current: " + String(BAUD_NAMES[currentIndex]) + "   ");
        printLCD(0, 3, "O:Save  E:Cancel");
        
        key = getKey();
        if (key == KEY_UP) {
            if (currentIndex > 0) currentIndex--;
        } else if (key == KEY_DOWN) {
            if (currentIndex < BAUD_COUNT - 1) currentIndex++;
        } else if (key == KEY_ENTER) {
            serialSettings.baudRateIndex = currentIndex;
            printLCD(0, 2, "Saved!            ");
            delay(1000);
            break;
        } else if (key == KEY_ESCAPE) {
            break;
        }
        delay(100);
    }
    
    displayMenu();
}

void setDataBits() {
    clearLCD();
    printLCD(0, 0, "Set Data Bits");
    printLCD(0, 1, "Use U/D to select");
    
    char key;
    uint8_t currentIndex = serialSettings.dataBitsIndex;
    
    while (true) {
        // Display current selection
        printLCD(0, 2, "Current: " + String(DATA_BITS_OPTIONS[currentIndex]) + "   ");
        printLCD(0, 3, "O:Save  E:Cancel");
        
        key = getKey();
        if (key == KEY_UP) {
            if (currentIndex > 0) currentIndex--;
        } else if (key == KEY_DOWN) {
            if (currentIndex < DATA_BITS_COUNT - 1) currentIndex++;
        } else if (key == KEY_ENTER) {
            serialSettings.dataBitsIndex = currentIndex;
            printLCD(0, 2, "Saved!            ");
            delay(1000);
            break;
        } else if (key == KEY_ESCAPE) {
            break;
        }
        delay(100);
    }
    
    displayMenu();
}

void setParity() {
    clearLCD();
    printLCD(0, 0, "Set Parity");
    printLCD(0, 1, "Use U/D to select");
    
    char key;
    uint8_t currentIndex = serialSettings.parityIndex;
    
    while (true) {
        // Display current selection
        printLCD(0, 2, "Current: " + String(PARITY_NAMES[currentIndex]) + "   ");
        printLCD(0, 3, "O:Save  E:Cancel");
        
        key = getKey();
        if (key == KEY_UP) {
            if (currentIndex > 0) currentIndex--;
        } else if (key == KEY_DOWN) {
            if (currentIndex < PARITY_COUNT - 1) currentIndex++;
        } else if (key == KEY_ENTER) {
            serialSettings.parityIndex = currentIndex;
            printLCD(0, 2, "Saved!            ");
            delay(1000);
            break;
        } else if (key == KEY_ESCAPE) {
            break;
        }
        delay(100);
    }
    
    displayMenu();
}

void setStopBits() {
    clearLCD();
    printLCD(0, 0, "Set Stop Bits");
    printLCD(0, 1, "Use U/D to select");
    
    char key;
    uint8_t currentIndex = serialSettings.stopBitsIndex;
    
    while (true) {
        // Display current selection
        printLCD(0, 2, "Current: " + String(STOP_BITS_NAMES[currentIndex]) + "   ");
        printLCD(0, 3, "O:Save  E:Cancel");
        
        key = getKey();
        if (key == KEY_UP) {
            if (currentIndex > 0) currentIndex--;
        } else if (key == KEY_DOWN) {
            if (currentIndex < STOP_BITS_COUNT - 1) currentIndex++;
        } else if (key == KEY_ENTER) {
            serialSettings.stopBitsIndex = currentIndex;
            printLCD(0, 2, "Saved!            ");
            delay(1000);
            break;
        } else if (key == KEY_ESCAPE) {
            break;
        }
        delay(100);
    }
    
    displayMenu();
}

void setFlowControl() {
    clearLCD();
    printLCD(0, 0, "Flow Control");
    printLCD(0, 1, "Use U/D to select");
    
    char key;
    bool currentSetting = serialSettings.flowControlEnabled;
    
    while (true) {
        // Display current selection
        printLCD(0, 2, "Current: RTS/CTS " + String(currentSetting ? "ON " : "OFF"));
        printLCD(0, 3, "O:Save  E:Cancel");
        
        key = getKey();
        if (key == KEY_UP || key == KEY_DOWN) {
            currentSetting = !currentSetting;
        } else if (key == KEY_ENTER) {
            serialSettings.flowControlEnabled = currentSetting;
            printLCD(0, 2, "Saved!            ");
            delay(1000);
            break;
        } else if (key == KEY_ESCAPE) {
            break;
        }
        delay(100);
    }
    
    displayMenu();
}

void saveSerialSettings() {
    clearLCD();
    printLCD(0, 0, "Save Settings");
    printLCD(0, 1, "Saving to EEPROM...");
    
    // Call the function to save settings to EEPROM
    saveSettingsToEEPROM();
    
    delay(1000);
    printLCD(0, 2, "Settings saved!");
    printLCD(0, 3, "Any key to return");
    
    // Wait for key press
    while (getKey() == NO_KEY) {
        delay(100);
    }
    
    displayMenu();
}